# A parser for mate formatted nested dict values of python modules, classes and functions


from argparse import Namespace
import re
from typing import Callable
from .utils import utils
from .utils.bunch import Bunch
import ipdb


def parse_py_module_to_dict(
    definition: dict,
    object,
    generate_defauls: bool = False,
):

    # type annotation
    if "self" in definition and definition["self"] == True:
        return definition

    merged_params, error = utils.get_function_parameters(
        object, definition, generate_defauls
    )
    definition["params"] = merged_params.copy()

    return definition, error


object_dict = {}

# creates a module object from a dict and calls it with the params
def parse_module_object(
    root: Bunch,
    object: Bunch,
    base_module: str = "",
    root_module: str = "",
    map_key_values: dict = {},
):

    module_class, params = parse_module_class_recursive(
        root, object, base_module, root_module, map_key_values
    )

    # ipdb.set_trace()
    if "self" in object.keys() and object["self"] == True:
        return module_class

    for key, value in params.items():
        if type(value) == dict and "object_key" in value and not "module" in value:
            # ipdb.set_trace()
            # find the object from the root
            if "function_call" in value:
                function = value["function_call"]
                function = getattr(object_dict[value["object_key"]], function)
                py_object = function(value["object_key"])
                # object = function(value[obj])
            else:
                py_object = object_dict[value["object_key"]]
            params[key] = py_object

    if "function" in object.keys():
        function = getattr(module_class, object["function"])
        py_object = function(**params)
    elif "class" in object.keys():
        # ipdb.set_trace()
        py_object = module_class(**params)

    # save object in dict for later use
    if "object_key" in object and "module" in object.keys():
        # ipdb.set_trace()
        object_dict[object["object_key"]] = py_object

    return py_object

    # function_name = object["function"] if "function" in object.keys() else "__init__"
    # function = getattr(module_class, function_name)
    return module_class(**params)


# Loads a module from a dict, importing hiearchy is local -> base -> global
def load_module(object: Bunch, base_module: str, root_module: str):

    assert "module" in object
    assert "class" or "function" in object
    fromlist = [object["class"]] if "class" in object else [object["function"]]
    try:
        module = __import__(base_module + "." + object["module"], fromlist=fromlist)
    except ModuleNotFoundError:
        # now try shared imports
        try:
            module = __import__(root_module + "." + object["module"], fromlist=fromlist)
        except ModuleNotFoundError:

            # lastly try global imports
            try:
                module = __import__(object["module"], fromlist=fromlist)
            except ModuleNotFoundError:
                return object

    if "class" in object:
        module = getattr(module, object["class"])

    return module


def __parse_dict_object(
    root: Bunch,
    object: dict,
    base_module: str,
    root_module: str,
    map_key_values: dict,
):

    if type(object) == dict:

        if "module" and ("class" or "function") in object.keys():
            parsed_module = parse_module_object(
                root, object, base_module, root_module, map_key_values
            )
            return parsed_module

        # check if the object is a dict with a function
        for key, value in object.items():
            object[key] = __parse_dict_object(
                root, value, base_module, root_module, map_key_values
            )
            # if type(value) == dict and (
            #     "module" and ("class" or "function") in value.keys()
            # ):
            #     object[key] = parse_module_object(
            #         root, value, base_module, root_module, map_key_values
            #     )

            # elif type(value) == dict:
            #     for _key, _value in value.items():
            #         if type(_value) == dict and (
            #             "module" and ("class" or "function") in _value.keys()
            #         ):
            #             object[key][_key] = parse_module_object(
            #                 root, _value, base_module, root_module, map_key_values
            #             )

            # if type(value) == list:
            #     for index, item in enumerate(value):
            #         if type(item) == dict and (
            #             "module" and ("class" or "function") in item.keys()
            #         ):
            #             object[key][index] = parse_module_object(
            #                 root, item, base_module, root_module, map_key_values
            #             )

    if type(object) == list:
        result = [
            __parse_dict_object(root, item, base_module, root_module, map_key_values)
            for item in object
        ]
        return result

    return object


# Function to generate children of a object
def __generate_children(
    root: Bunch,
    object: Bunch,
    base_module: str,
    root_module: str,
    map_key_values: dict = {},
):

    try:
        if "params" in object:
            params = object.get("params", {}).copy()
    except KeyError:
        params = {}
        object["params"] = {}
        # either empty params, no params or just a type?

    # recursively check if module has a object child
    # parse it and add it to the params

    new_params = {}
    # parse params
    for key, value in object["params"].items():
        if type(value) == dict or type(value) == list:
            new_params[key] = __parse_dict_object(
                root, value, base_module, root_module, map_key_values
            )

    params.update(new_params)

    # replace {regex} with values from map_key_values
    for key, value in params.items():

        # regex for {anystring}
        if type(value) == str and re.search(r"{.*}", value):
            # get the key name
            key_name = re.search(r"{.*}", value).group(0)[1:-1]
            # replace the value
            params[key] = value.replace("{" + key_name + "}", map_key_values[key_name])

    return params


# Function to parse a nested dict of python modules, classes and functions
def parse_module_class_recursive(
    root: Bunch,
    object: dict,
    base_module: str = "",
    root_module: str = "",
    map_key_values: dict = {},
):

    module = load_module(object, base_module, root_module)

    if "self" in object:
        return module, {}

    params = __generate_children(root, object, base_module, root_module, map_key_values)

    return module, params


def generate_params_no_depth(
    object: Bunch,
    base_module: str,
    root_module: str,
    generate_defauls: bool = False,
):

    if "self" in object and object["self"] == True:
        return object, None

    if "module" not in object:
        return object, None

    module = load_module(object, base_module, root_module)

    callable = object["function"] if "function" in object else "__init__"

    object_definition, error = parse_py_module_to_dict(
        object, getattr(module, callable), generate_defauls
    )

    return object_definition, error


def generate_params_recursively(
    object: Bunch,
    base_module: str,
    root_module: str,
    generate_default_params: bool = False,
):

    if "self" in object and object["self"] == True:
        return object, None

    assert "module" in object

    object, error = generate_params_no_depth(
        object, base_module, root_module, generate_default_params
    )
    errors = [error] if error else []

    if "params" in object:
        for key, value in object["params"].items():
            if type(value) == dict and (
                "module" and ("class" or "function") in value.keys()
            ):
                # ipdb.set_trace()
                obj_def, error = generate_params_recursively(
                    value, base_module, root_module, generate_default_params
                )
                object["params"][key] = obj_def.copy()
                errors += [error] if error else []

            if type(value) == list:

                for index, item in enumerate(value):
                    if type(item) == dict and (
                        "module" and ("class" or "function") in item.keys()
                    ):

                        obj_def, error = generate_params_recursively(
                            item, base_module, root_module, generate_default_params
                        )
                        object["params"][key][index] = obj_def.copy()
                        errors += [error] if error else []

    return object, errors


def generate_params(
    root: Bunch,
    base_module: str,
    root_module: str,
    generate_defauls: bool = False,
):

    # params = root.copy()

    errors = []
    root = root.copy()

    # # first generate the params for the root
    # if "module" in root:
    #     root, error = generate_params_no_depth(root, base_module, root_module)
    #     errors += [error] if error else []

    # then generate the params for the children
    for key, value in root.items():
        if type(value) == dict and (
            "module" and ("class" or "function") in value.keys()
        ):
            gen_def, err = generate_params_recursively(
                root[key], base_module, root_module, generate_defauls
            )
            root[key] = gen_def.copy()

            errors += err

    # remove None from erros
    errors = [error for error in errors if error]

    return root, errors


def load_python_object(
    object: Bunch,
    root: Bunch,
    base_module: str,
    root_module: str,
    map_key_values: dict = {},
):

    # _, _ = generate_full_params(root, base_module, root_module)

    if "module" in object:
        model_class, obj_params = parse_module_class_recursive(
            root,
            object,
            base_module=base_module,
            root_module=root_module,
            map_key_values=map_key_values,
        )

        return model_class(**obj_params)

    result = {}

    for key, value in object.items():
        if type(value) == dict and (
            "module" and ("class" or "function") in value.keys()
        ):
            result[key] = load_python_object(
                value, root, base_module, root_module, map_key_values
            )
        elif type(value) == list:
            result[key] = [
                load_python_object(item, root, base_module, root_module, map_key_values)
                for item in value
            ]
        elif type(value) == dict:
            result[key] = load_python_object(
                value, root, base_module, root_module, map_key_values
            )
        else:
            result[key] = value

    return result


if __name__ == "__main__":
    pass



from typing import Callable
import inspect
import ipdb


def get_function_args(
    function: Callable,
    args={},
    generate_defaults: bool = True,
    generate_none: bool = False,
):

    params = args.to_dict() if hasattr(args, "to_dict") else args
    default = {}
    errors = []

    for param_name, param in inspect.signature(function).parameters.items():

        # only add parameters that are not self or exist in the params
        if param_name in params or param_name == "self":
            continue

        # only allow named parameters and not *args or **kwargs
        if param.kind != param.POSITIONAL_OR_KEYWORD:
            continue

        if param.default is not param.empty:
            # ignore None values
            if param.default != None and generate_defaults:
                default[param_name] = param.default
            elif generate_none:
                default[param_name] = None

        elif param.annotation is not param.empty:
            default[param_name] = f"Fix me! {param.annotation}"
            error = f"Missing parameter {param_name} for {object['module']}\nHint: {param.annotation}"
            errors.append(error)
        else:
            default[param_name] = "Fix me!"
            error = f"Missing parameter {param_name} for {object['module']}\nHint: Add a default value or type annotation"
            errors.append(error)

    params.update(default)

    return params, errors



# Bambilla, API for bamiblla json format and python objects

from .node import Node, NodeDict
import ipdb


class Bombilla:
    def __init__(
        self,
        bambilla_dict: dict,
        root_module: str = "",
        base_module: str = "",
        object_key_map: dict = {},
    ) -> None:

        assert isinstance(bambilla_dict, dict), "Bambilla must be a dict"

        Node.set_config(base_module, root_module, object_key_map)
        self.root_node = NodeDict(dict(bambilla_dict))

    def load(self):
        self.root_node.__load__()

    def execute(self):

        self.root_node.__call__()

    def generate_full_dict(self):
        self.root_node.__load__()

        return self.root_node.generate_full_dict()

    def execute_method(
        self, method_name: str, object_name: str = None, *args, **kwargs
    ):

        self.root_node.find(object_name).call_method(method_name, *args, **kwargs)

    def find(self, object_name: str):
        return Node._key_value_map[object_name]



from distutils import errors
from re import S
from typing import Optional, Any, Callable, Type, Union
import json

from . import utils
import regex as re

import ipdb

SimpleType = Union[str, int, float, bool, None]


class Node:
    object_key: Optional[str] = None
    _py_object: Optional[object] = None

    def __init__(self, args, parent=None, **kawrgs) -> None:
        self._original_keys = args.keys()
        self._original_args = dict(args)
        self._uno_key_value_map = {}
        self._parent = parent

        if args is None:
            args = self._json()
        for key, val in args.items():
            setattr(self, key, val)

    def _get_python_object(self) -> object:
        pass

    def find(self):
        pass

    @staticmethod
    def set_config(base_module: str, root_module: str, key_value_map: dict):
        Node._base_module = base_module
        Node._root_module = root_module
        Node._key_value_map = key_value_map

    def load_dynamic_objects(self):
        for key, value in self.__dict__.items():
            if (
                type(value) == str
                and key in self._original_keys
                and re.search(r"{.*}", value)
            ):
                # get the key name
                key_name = re.search(r"{.*}", value).group(0)[1:-1]

                dynamic_object = Node._key_value_map[key_name]
                self._uno_key_value_map[key] = value

                setattr(self, key, dynamic_object)

            if type(value) == list:
                set = False
                for i, item in enumerate(value):
                    if (
                        type(item) == str
                        and key in self._original_keys
                        and re.search(r"{.*}", item)
                    ):
                        # get the key name
                        key_name = re.search(r"{.*}", item).group(0)[1:-1]

                        dynamic_object = Node._key_value_map[key_name]
                        value[i] = dynamic_object
                        set = True

                if set:

                    self._uno_key_value_map[key] = value
                    setattr(self, key, value)

    def post_object_creation(self):
        # ipdb.set_trace()
        if "object_key" in self._original_keys:
            Node._key_value_map[self.object_key] = self._py_object

    def load_module(self):
        assert "module" in self.__dict__, "module not found"

        fromlist = [self.class_name] if hasattr(self, "class_name") else []
        fromlist = [getattr(self, "class")] if hasattr(self, "class") else fromlist

        if hasattr(self, "function") and not hasattr(self, "class_name"):
            fromlist = [self.function]

        module_list = [
            Node._base_module + "." + self.module,
            Node._root_module + "." + self.module,
            self.module,
        ]
        module = None
        for module_name in module_list:
            # print(f"Trying to import {module_name}")
            try:
                module = __import__(module_name, fromlist=fromlist)
                break
            except ModuleNotFoundError as err:
                # print(f"{err.name=} {err.path=} {err.msg=}")
                # print(f"{module_name=}")
                if (
                    err.name != None
                    and err.name.split(".")[0] != module_name.split(".")[0]
                ):
                    raise err
                pass

        if module == None:
            raise ModuleNotFoundError(f'Module "{self.module}" not found')

        if "class_name" in self.__dict__:
            # ipdb.set_trace()
            module = getattr(module, self.class_name)
        # if "class_name" in self:
        #     module = getattr(module, getattr[self, "class_name"])

        return module

    def _json(self):
        # returns a simple json representing the node (filtering out private properties and methods)

        def to_json(obj):
            if isinstance(obj, Node):
                return {
                    key: to_json(val)
                    for key, val in obj.__dict__.items()
                    if not key.startswith("_")
                    and not callable(val)
                    and type(val) in [str, int, float, bool, dict]
                }
            elif isinstance(obj, list):
                return [to_json(item) for item in obj]
            elif isinstance(obj, dict):
                return {key: to_json(val) for key, val in obj.items()}
            else:
                return obj

        return to_json(self)

    def __str__(self):
        return json.dumps(self._json(), indent=4)

    def __repr__(self):
        return str(self)

    @classmethod
    def _base_json(cls):
        # returns a simple json representing the node (filtering out private properties and methods)
        cur_params = {
            key: val
            for key, val in cls.__dict__.items()
            if not key.startswith("_") and not callable(val) and type(val) != object
        }
        for base_class in cls.__bases__:
            if base_class != object:
                cur_params |= {
                    key: val
                    for key, val in base_class.__dict__.items()
                    if not key.startswith("_")
                    and not callable(val)
                    and not (hasattr(val, "__get__") and callable(val.__get__))
                }
        return cur_params

    @classmethod
    def is_one(cls: Type, obj: Any) -> bool:
        # checks that the given object has the same property as the current node
        """
        return isinstance(obj, dict) and (
            sorted(list(obj.keys())) == node_properties
        )
        """
        if not isinstance(obj, dict):
            return False
        for key, val in cls._base_json().items():
            if key not in obj and val is not None:
                return False
        return True

    @classmethod
    def assert_is_one(cls: Type, obj: Any):
        if not cls.is_one(obj):
            raise SyntaxError(
                f"Object {obj} is not a valid node Node. \n This is how it should look like:\n{cls._base_json()}"
            )

    def __load__(self, *args, **kwargs):
        pass

    def __call__(self, *args, **kwargs):

        if self._py_object != None:
            return self._py_object
        else:

            super_dict = self.__dict__
            super_dict = dict(
                {
                    key: val() if isinstance(val, Node) else val
                    for key, val in super_dict.items()
                    if not key.startswith("_") and key in self._original_keys
                }
            )
            return super_dict

    # def __dict__(self):
    #     super_dict = super().__dict__
    #     return dict({key: val for key, val in super_dict.items() if not key.startswith("_")})


class NodeDict(Node):
    def __init__(self, args, **kawrgs) -> None:
        # ipdb.set_trace()
        assert type(args) == dict, "args must be a dict"
        super().__init__(args, **kawrgs)
        self._node_key_dict = {}

    def __load__(self, parent: Optional[Node] = None):

        # ipdb.set_trace()
        # self._parent = parent
        for key, val in self.__dict__.items():

            if not key.startswith("_") and key in self._original_keys:
                val = load_node(val, parent=self)
                if isinstance(val, Node):
                    val.__load__(self)
                    setattr(self, key, val)
                    self._node_key_dict[key] = val

                elif isinstance(val, list):
                    nodes = []
                    for i, item in enumerate(val):
                        item = load_node(item, parent=self)
                        if isinstance(item, Node):
                            item.__load__(self)
                            nodes.append(item)
                            val[i] = item
                            self._node_key_dict[key] = nodes
                    setattr(self, key, val)

        return self

    def __call__(self, *args: Any, **kwds: Any) -> Any:

        self.load_dynamic_objects()

        def val_to_call(val):
            if isinstance(val, Node):
                return val()
            elif isinstance(val, list):
                return [val_to_call(item) for item in val]
            else:
                return val

        result = {
            key: val_to_call(val)
            for key, val in self.__dict__.items()
            if not key.startswith("_") and key in self._original_keys
        }
        return result

    def find(self, key_name):
        assert (
            self._node_key_dict != None
        ), "node_key_dict is None, cannot find before loading"
        return self._node_key_dict[key_name]

    def to_dict(self):

        if self._node_key_dict == None:
            self.__load__()

        def load_node(node):
            if isinstance(node, Node):
                return node.to_dict()
            elif isinstance(node, list):
                return [load_node(item) for item in node]
            else:
                return node

        result = {}

        for key, val in self.__dict__.items():
            if not key.startswith("_") and key in self._original_keys:
                if key in self._node_key_dict:
                    val = load_node(self._node_key_dict[key])
                    result[key] = val
                elif key in self._uno_key_value_map:
                    result[key] = self._uno_key_value_map[key]

                elif type(val) == object:
                    result[key] = self._original_args[key]
                else:
                    result[key] = val

        return result

    def generate_full_dict(self):

        if self._node_key_dict == None:
            self.__load__()

        errors = []

        def load_node(node, errors=errors):
            if isinstance(node, Node):
                n, e = node.generate_full_dict()
                errors += e
                return n
            elif isinstance(node, list):
                for i, item in enumerate(node):
                    if isinstance(item, Node):
                        n, e = item.generate_full_dict()
                        errors += [e]
                        node[i] = n

                return node

            else:
                return node

        result = {}

        for key, val in self.__dict__.items():
            if not key.startswith("_") and key in self._original_keys:

                if key in self._uno_key_value_map:
                    result[key] = self._uno_key_value_map[key]
                elif key in self._node_key_dict:
                    val = load_node(self._node_key_dict[key])
                    result[key] = val

                elif type(val) == list or type(val) == Node:
                    result[key] = load_node(val)
                else:
                    result[key] = val

        errors = [e for e in errors if e != None and e != []]

        return result, errors

    def parse_params(self):

        if self._node_key_dict == None:
            self.__load__()

        errors = []
        params = {}

        for key, val in self.__dict__.items():

            if not key.startswith("_") and key in self._original_keys:

                if key in self._node_key_dict:
                    p, e = self._node_key_dict[key].parse_params()
                    params[key] = p
                    errors += e
                elif isinstance(val, Node):
                    p, e = val.parse_params()
                    params[key] = p
                    errors += e

                else:
                    params[key] = val

        # remove None from errors
        errors = [e for e in errors if e != None and e != []]

        return params, errors


class ObjectReference(Node):
    reference_key: str = ""
    _reference: Optional[Node] = None

    def __init__(self, args, parent: Optional[Node] = None, **kawrgs) -> None:
        super().__init__(args, parent=parent, **kawrgs)

    def __call__(self, *args: Any, **kwds: Any) -> Any:
        return Node._key_value_map[self.reference_key]


class MethodCall(Node):
    function_call: str = ""
    reference_key: Optional[str] = None
    params: dict = {}

    def __load__(self, parent=None, *args, **kwargs):

        self.param_node = NodeDict(self.params, parent=self)
        self.param_node.__load__(self)
        return self

    def __call__(self, parent: Optional[object] = None, *args, **kwargs):

        if self._py_object != None:
            return self._py_object
        # first create DictNode of params
        params = self.param_node()
        # print("method call params", params)

        # then call the function
        object = Node._key_value_map[self.reference_key]
        function = getattr(object, self.function_call)
        self._py_object = function(**params)
        self.post_object_creation()
        return self._py_object

    def __init__(self, args, parent: Optional[object] = None):
        super().__init__(args)

    def to_dict(self):
        return {
            "reference_key": self.reference_key,
            "function_call": self.function_call,
            "params": self.param_node.to_dict(),
        }

    def generate_full_dict(self):

        # ipdb.set_trace()
        full_params, errors = self.param_node.parse_params()
        errors = [e for e in errors if e != None and e != []]
        return {
            "reference_key": self.reference_key,
            "function_call": self.function_call,
            "params": full_params,
        }, errors

    def parse_params(self):

        function = getattr(Node._key_value_map[self.reference_key], self.function_call)

        # get the function args that are missing from the params
        params, errors = utils.get_function_args(function, self.param_node)
        errors = [e for e in errors if e != None and e != []]

        return params, errors


# Methodcall for objects
class ObjectMethodCall(Node):
    function: str = ""
    params: dict = {}

    def __init__(self, args, parent=None, **kawrgs) -> None:
        super().__init__(args, parent, **kawrgs)
        self._node = None

    def __load__(self, parent=None) -> object:

        if self._node == None:
            self._node = NodeDict(self.params, parent=self)
            self._node.__load__(self)
            return self
        # self._node = NodeDict(self.params)
        return self

    def __call__(self, *args, **kwargs):
        return self._node()

    def to_dict(self):
        return {"function": self.function, "params": self._node.to_dict()}

    def generate_full_dict(self):
        full_params, errors = self._node.generate_full_dict()
        return {"function": self.function, "params": full_params}, errors


class FunctionModuleCall(Node):
    function: str = ""
    module: str = ""
    params: dict = {}
    method_args: Optional[list[ObjectMethodCall]] = None

    def __load__(self, parent=None) -> object:

        self._param_node = NodeDict(self.params)
        self._param_node.__load__(self)
        self._method_args_nodes = (
            [ObjectMethodCall(m, self) for m in self.method_args]
            if self.method_args
            else []
        )
        return self

    def to_dict(self):
        return {
            "function": self.function,
            "module": self.module,
            "params": self._param_node.to_dict(),
        }

    def generate_full_dict(self):
        result = {
            "function": self.function,
            "module": self.module,
        }
        errs = []
        if self.method_args != None:
            full_method_args, err = self.parse_method_args()
            result["method_args"] = full_method_args
            errs += err

        full_params, err = self.parse_params()
        result["params"] = full_params

        errs += err

        # remove None from errors
        errs = [e for e in errs if e != None and e != []]

        return result, errs

    def parse_params(self):

        module = self.load_module()
        function = getattr(module, self.function)

        params, errors = utils.get_function_args(function, self._param_node)

        errors = [e for e in errors if e != None and e != []]

        return params, errors

    def __call__(self):

        params = self._param_node()

        module = self.load_module()

        function = getattr(module, self.function)

        _arg, _kwarg = flatten_nameless_params(params)

        if _arg:
            self._py_object = function(*_arg, **_kwarg)
        else:
            self._py_object = function(**params)
        self.post_object_creation()
        return self._py_object

    def call_method(self, method_name: str, *args, **kwargs):
        method_arg_names = [arg["function"] for arg in self.method_args]
        assert method_name in method_arg_names

        idx = method_arg_names.index(method_name)
        method_args = self._method_args_nodes[idx]
        # ipdb.set_trace()

        # method_args =  #load_node(method_args, parent=self)
        # ipdb.set_trace()
        method_args.__load__(self)
        method_args = method_args()
        f, p = flatten_nameless_params(method_args)

        method = getattr(self._py_object, method_name)

        if f:
            return method(*args, *f, **p, **kwargs)

        return method(*args, **p, **kwargs)

    def parse_method_args(self):

        assert self._py_object != None, "Object not created yet"
        assert self.method_args != None, "No method args"

        errors = []
        params = {}
        results = []

        for method in self._method_args_nodes:
            # ipdb.set_trace()

            function = method.function
            method.__load__(self)
            full_args, errs = method.generate_full_dict()
            errors += errs

            m = getattr(self._py_object, function)
            p, error = utils.get_function_args(m, full_args["params"])

            gen_method = {
                "function": function,
                "params": p,
            }

            results += [gen_method]
            errors += error

        # remove None from errors
        errors = [e for e in errors if e != None and e != []]

        return results, errors


def flatten_nameless_params(params: dict) -> dict:

    flat = params.pop("", None)
    flats = []
    while flat != None:
        flats.append(flat)
        flat = params.pop("", None)

    return flats, params


class Object(Node):
    module: str = ""
    class_name: str = ""
    params: dict = {}  # param is actualy a dictnode
    method_args: Optional[list[ObjectMethodCall]] = None

    def __load__(self, parent: Optional[object] = None) -> object:

        if self._py_object != None:
            return self

        self.param_node = NodeDict(self.params)

        self.param_node.__load__()

        self._method_args_nodes = (
            [ObjectMethodCall(m, self) for m in self.method_args]
            if self.method_args
            else []
        )

        return self

    def __call__(self, *args, **kwargs):

        if self._py_object != None:
            return self._py_object

        module = self.load_module()

        self._py_object = module(**self.param_node())
        self.post_object_creation()
        return self._py_object

    def to_dict(self):
        return {
            "module": self.module,
            "class_name": self.class_name,
            "params": self.param_node.to_dict(),
        }

    def generate_full_dict(self):
        result = {
            "module": self.module,
            "class_name": self.class_name,
        }
        errs = []
        if self.method_args != None:
            full_method_args, err = self.parse_method_args()
            result["method_args"] = full_method_args
            errs += [err]

        full_params, err = self.parse_params()
        result["params"] = full_params

        errs += err

        # remove None from errors
        errs = [e for e in errs if e != None]

        return result, errs

    def parse_params(self):

        module = self.load_module()

        params, err = self.param_node.generate_full_dict()

        # ipdb.set_trace()
        errors = err

        params, erros = utils.get_function_args(module, params)

        errors += erros
        errors = [e for e in errors if e != None and e != []]

        return params, errors

    def __init__(self, args: Optional[dict], parent: Optional[object] = None):
        super().__init__(args, parent=parent)

    def call_method(self, method_name: str, *args, **kwargs):

        method_arg_names = [arg["function"] for arg in self.method_args]
        assert method_name in method_arg_names

        idx = method_arg_names.index(method_name)
        method_args = self.method_args[idx]["params"]

        method_args = load_node(method_args, parent=self)
        method_args.__load__(self)
        method_args = method_args()

        method = getattr(self._py_object, method_name)
        return method(*args, **method_args, **kwargs)

    def parse_method_args(self):

        # assert self._py_object != None, "Object not created yet"
        assert self.method_args != None, "No method args"

        errors = []
        params = {}
        results = []

        for method in self._method_args_nodes:
            # ipdb.set_trace()

            function = method.function
            method.__load__(self)
            full_args, errs = method.generate_full_dict()
            errors += errs

            m = getattr(self._py_object, function)
            # ipdb.set_trace()
            p, error = utils.get_function_args(m, full_args["params"])

            gen_method = {
                "function": function,
                "params": p,
            }

            results += [gen_method]
            errors += error

        # remove None from errors
        errors = [e for e in errors if e != None and e != []]

        return results, errors


class AnnonymousObject(Node):
    def __init__(self, args: Optional[dict] = None) -> None:
        super().__init__(args)
        setattr(self, "", "")


node_types: list[Type] = [
    Object,
    MethodCall,
    FunctionModuleCall,
    ObjectMethodCall,
    ObjectReference,
    NodeDict,
]
SyntaxNode = Union[
    Object, FunctionModuleCall, MethodCall, ObjectReference, ObjectMethodCall, NodeDict
]


def load_node(args, parent=None):
    if isinstance(args, dict):
        # ipdb.set_trace()
        node: Optional[SyntaxNode] = None
        for node_type in node_types:
            if node_type.is_one(args):
                node = node_type(args, parent=parent)
                return node

    return args



from setuptools import setup, find_packages

setup(
    name="bomba",
    description="A cowsay clone for python in one file.",
    author="Giulio Zani, Ali Rahimi",
    author_email="yerba.mate.dl@proton.me",
    url="https://github.com/ilex-paraguariensis/bombilla",
    python_requires=">=3.9",
    version="1.1.112",
    packages=find_packages("packages", exclude=["tests", "examples"]),
    include_package_data=True,
    package_dir={"": "packages/"},
    license="Apache License 2.0",
    license_files=("LICENSE.md",),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
    ],
    scripts=["./src/bomba"],
)
